#!/usr/bin/env python3

#
# NOSA HEADER START
#
# The contents of this file are subject to the terms of the NASA Open
# Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may
# not use this file except in compliance with the Agreement.
#
# You can obtain a copy of the agreement at
#   docs/NASA_Open_Source_Agreement_1.3.txt
# or
#   https://cdaweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
#
# See the Agreement for the specific language governing permissions
# and limitations under the Agreement.
#
# When distributing Covered Code, include this NOSA HEADER in each
# file and include the Agreement file at
# docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the
# following below this NOSA HEADER, with the fields enclosed by
# brackets "[]" replaced with your own identifying information:
# Portions Copyright [yyyy] [name of copyright owner]
#
# NOSA HEADER END
#
# Copyright (c) 2023 United States Government as represented by
# the National Aeronautics and Space Administration. No copyright is
# claimed in the United States under Title 17, U.S.Code. All Other
# Rights Reserved.
#

""" 
Module for accessing the Heliophysics Data Portal (HDP) web services
https://heliophysicsdata.gsfc.nasa.gov/WebServices/.
"""

import platform
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import ParseError
import logging
from typing import Dict, List, Union
import requests

from hdpws import __version__, NAMESPACES
from hdpws.dateinterval import DateInterval
from hdpws.resourcetype import ResourceType



class HdpWs:
    """
    Class representing the web service interface to NASA's
    Heliophysics Data Portal (HDP) <https://heliophysicsdata.gsfc.nasa.gov/>.

    Parameters
    ----------
    endpoint
        URL of the HDP web service.  If None, the default is
        'https://heliophysicsdata.gsfc.nasa.gov/WS/hdp/1/'.
    timeout
        Number of seconds to wait for a response from the server.
    proxy
        HTTP proxy information.  For example,<pre>
        proxies = {
          'http': 'http://10.10.1.10:3128',
          'https': 'http://10.10.1.10:1080',
        }</pre>
        Proxy information can also be set with environment variables.
        For example,<pre>
        $ export HTTP_PROXY="http://10.10.1.10:3128"
        $ export HTTPS_PROXY="http://10.10.1.10:1080"</pre>
    ca_certs
        Path to certificate authority (CA) certificates that will
        override the default bundle.
    disable_ssl_certificate_validation
        Flag indicating whether to validate the SSL certificate.
    user_agent
        A value that is appended to the HTTP User-Agent values.

    Notes
    -----
    The logger used by this class has the class' name (HdpWs).  By default,
    it is configured with a NullHandler.  Users of this class may configure
    the logger to aid in diagnosing problems.

    This class is dependent upon xml.etree.ElementTree module which is
    vulnerable to an "exponential entity expansion" and "quadratic blowup
    entity expansion" XML attack.  However, this class only receives XML
    from the (trusted) HDP server so these attacks are not a threat.  See
    the xml.etree.ElementTree "XML vulnerabilities" documentation for
    more details
    <https://docs.python.org/3/library/xml.html#xml-vulnerabilities>.
    """
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    def __init__(
            self,
            endpoint=None,
            timeout=None,
            proxy=None,
            ca_certs=None,
            disable_ssl_certificate_validation=False,
            user_agent=None):

        self.logger = logging.getLogger(type(self).__name__)
        self.logger.addHandler(logging.NullHandler())

        self.retry_after_time = None

        self.logger.debug('endpoint = %s', endpoint)
        self.logger.debug('ca_certs = %s', ca_certs)
        self.logger.debug('disable_ssl_certificate_validation = %s',
                          disable_ssl_certificate_validation)

        if endpoint is None:
            self._endpoint = 'https://heliophysicsdata.gsfc.nasa.gov/WS/hdp/1/'
        else:
            self._endpoint = endpoint

        self._user_agent = 'hdpws/' + __version__ + ' (' + \
            platform.python_implementation() + ' ' \
            + platform.python_version() + '; '+ platform.platform() + ')'

        if user_agent is not None:
            self._user_agent += ' (' + user_agent + ')'

        self._request_headers = {
            'Content-Type' : 'application/xml',
            'Accept' : 'application/xml',
            'User-Agent' : self._user_agent
        }
        self._session = requests.Session()
        #self._session.max_redirects = 0
        self._session.headers.update(self._request_headers)

        if ca_certs is not None:
            self._session.verify = ca_certs

        if disable_ssl_certificate_validation is True:
            self._session.verify = False

        if proxy is not None:
            self._proxy = proxy

        self._timeout = timeout

    # pylint: enable=too-many-arguments


    def __str__(self) -> str:
        """
        Produces a string representation of this object.

        Returns
        -------
        str
            A string representation of this object.
        """
        return 'HdpWs(endpoint=' + self._endpoint + ', timeout=' + \
               str(self._timeout) + ')'


    def __del__(self):
        """
        Destructor.  Closes all network connections.
        """

        self.close()


    def close(self) -> None:
        """
        Closes any persistent network connections.  Generally, deleting
        this object is sufficient and calling this method is unnecessary.
        """
        self._session.close()


    def get_instrument_ids(
            self
        ) -> Dict:
        """
        Gets all /Spase/Instrument/ResourceID values available at HDP.

        Returns
        -------
        Dict
            Dictionary containing a 'InstrumentID' key with a value
            of a List containing all /Spase/Instrument/ResourceID values
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_simple_resource('Spase/Instrument/ResourceID',
                                          'ResourceID', 
                                          'InstrumentID')


    def get_repository_ids(
            self
        ) -> Dict:
        """
        Gets all /Spase/Repository/ResourceID values available at HDP.

        Returns
        -------
        Dict
            Dictionary containing a 'RespositoryID' key with a value
            of a List containing all /Spase/Respository/ResourceID values
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_simple_resource('Spase/Repository/ResourceID',
                                          'ResourceID', 
                                          'RepositoryID')


    def get_measurement_types(
            self
        ) -> Dict:
        """
        Gets all /Spase/MeasurementType values available at HDP.

        Returns
        -------
        Dict
            Dictionary containing a 'MeasurementType' key with a value
            of a List containing all /Spase/MeasurementType values
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_simple_resource('Spase/MeasurementType',
                                          'MeasurementType', 
                                          'MeasurementType')


    def get_spectral_ranges(
            self
        ) -> Dict:
        """
        Gets all /Spase/SpectralRange values available at HDP.

        Returns
        -------
        Dict
            Dictionary containing a 'SpectralRange' key with a value
            of a List containing all /Spase/SpectralRange values
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_simple_resource('Spase/SpectralRange',
                                          'SpectralRange', 'SpectralRange')


    def get_phenomenon_types(
            self
        ) -> Dict:
        """
        Gets all /Spase/PhenomenonType values available at HDP.

        Returns
        -------
        Dict
            Dictionary containing a 'PhenomenonType' key with a value
            of a List containing all /Spase/PhenomenonType values
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_simple_resource('Spase/PhenomenonType',
                                          'PhenomenonType', 
                                          'PhenomenonType')


    def get_observatory_group_ids(
            self
        ) -> Dict:
        """
        Gets all /Spase/Observatory/ObservatoryGroupID values available
        at HDP.

        Returns
        -------
        Dict
            Dictionary containing a 'ObservatoryGroupID' key with a value
            of a List containing all /Spase/Observed/ObservatoryGroupID
            values with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_simple_resource('Spase/Observatory/ObservatoryGroupID',
                                          'ObservatoryGroupID', 'ObservatoryGroupID')


    def get_observatory_ids(
            self
        ) -> Dict:
        """
        Gets all /Spase/Observatory/ResourceID values available at HDP.

        Returns
        -------
        Dict
            Dictionary containing a 'ObservatoryID' key with a value
            of a List containing all /Spase/Observatory/ResourceID values
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_simple_resource('Spase/Observatory/ResourceID',
                                          'ResourceID', 'ObservatoryID')


    def get_observed_regions(
            self
        ) -> Dict:
        """
        Gets all /Spase/ObservedRegion values available at HDP.

        Returns
        -------
        Dict
            Dictionary containing a 'ObservedRegion' key with a value
            of a List containing all /Spase/ObservedRegion values
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_simple_resource('Spase/ObservedRegion',
                                          'Region', 'ObservedRegion')


    def get_spase(
            self,
            resource_ids: List[str]
        ) -> Dict:
        """
        Gets the specified SPASE documents from HDP.

        Parameters
        ----------
        resource_ids
            List of SPASE ResourceID values of the documents to get.

        Returns
        -------
            Dictionary containing a 'Result' key with a value of an
            ElementTree representation of the results as described by
            <https://heliophysicsdata.gsfc.nasa.gov/WebServices/hdpspase.xsd>
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        url = self._endpoint + 'Spase'

        query = {
            'ResourceID': resource_ids
        }
        response = self._session.get(url, params=query,
                                     timeout=self._timeout)

        self.logger.debug('response.url = %s', response.url)

        status = self.__get_status(response)
        if response.status_code != 200:
            return status

        result = {
            'Result': ET.fromstring(response.text)
        }

        result.update(status)
        return result


    def get_spase_data(
            self,
            resource_types: List[ResourceType],
            query: Dict,
            date_range: Union[List[str], DateInterval] = None
        ) -> Dict:
        """
        Gets the specified SPASE "data" documents from HDP.

        Parameters
        ----------
        resource_types
            List of SPASE ResourceTypes to get.  ResourceType.NUMERICAL_DATA
            , ResourceType.DISPLAY_DATA, or both.
        query
            Dictionary containing query parameters and values.  For
            example,<pre>
            query = { 
                'InstrumentID': 'spase://SMWG/Instrument/ACE/CRIS',
                'Cadence': '=PT1H',
                'ObservedRegion': 'Heliosphere.NearEarth',
                'MeasurementType': 'EnergeticParticles',
                'AccessRights': 'Open',
                'Format': 'CDF'
            }</pre>
        date_range
            A DateInterval or two element array of ISO 8601 string
            values of the start and stop date of the requested resources.
            for example,<pre>
            date_range = ['2022-01-01', '2022-01-02']
            </pre>

        Returns
        -------
            Dictionary containing a 'Result' key with a value of an
            ElementTree representation of the results as described by
            <https://heliophysicsdata.gsfc.nasa.gov/WebServices/hdpspase.xsd>
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_complex_resource(resource_types, query, date_range)


    def get_spase_catalog(
            self,
            query: Dict,
            date_range: Union[List[str], DateInterval] = None
        ) -> Dict:
        """
        Gets the specified SPASE Catalog documents from HDP.

        Parameters
        ----------
        query
            Dictionary containing query parameters and values.  For
            example,<pre>
            query = { 
                'InstrumentID': 'spase://SMWG/Instrument/ACE/MAG',
                'PhenomenonType': 'MagneticCloud',
                'Description': 'ICME'
            }</pre>
        date_range
            A DateInterval or two element array of ISO 8601 string
            values of the start and stop date of the requested resources.
            for example,<pre>
            date_range = ['1999-01-01', '1999-01-02']
            </pre>

        Returns
        -------
            Dictionary containing a 'Result' key with a value of an
            ElementTree representation of the results as described by
            <https://heliophysicsdata.gsfc.nasa.gov/WebServices/hdpspase.xsd>
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        return self.__get_complex_resource([ResourceType.CATALOG], query, date_range)


    def __get_simple_resource(
            self,
            resource: str,
            name: str,
            result_name: str
        ) -> Dict:
        """
        Gets all simple resource values available at HDP.

        Parameters
        ----------
        resource
            Resource path.  For example, Spase/ObservedRegion.
        name
            Element name to get from response.  For example, Region.
        result_name
            Name of key to return result in.  For example, 
            ObservedRegion.

        Returns
        -------
        Dict
            Dictionary containing a key with the given result_name and
            a value equal to a List of the given name element values from
            the response
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        url = self._endpoint + resource

        self.logger.debug('request url = %s', url)

        response = self._session.get(url, timeout=self._timeout)

        status = self.__get_status(response)
        if response.status_code != 200:
            return status

        mt_response = ET.fromstring(response.text)

        result = {
            result_name: []
        }

        for value in mt_response.findall('.//hdp:' + name,
                                         namespaces=NAMESPACES):

            result[result_name].append(value.text)

        result.update(status)
        return result


    def __get_complex_resource(
            self,
            resource_types: List[ResourceType],
            query: Dict,
            date_range: Union[List[str], DateInterval] = None
        ) -> Dict:
        """
        Gets the specified SPASE documents from HDP.

        Parameters
        ----------
        resource_types
            List of SPASE ResourceTypes to get.
        query
            Dictionary containing query parameters and values.
        date_range
            A DateInterval or two element array of ISO 8601 string
            values of the start and stop date of the requested resources.
            for example,<pre>
            date_range = ['2022-01-01', '2022-01-02']
            </pre>

        Returns
        -------
            Dictionary containing a 'Result' key with a value of an
            ElementTree representation of the results as described by
            <https://heliophysicsdata.gsfc.nasa.gov/WebServices/hdpspase.xsd>
            with the addition of the following key/values:<br>
            - HttpStatus: with the value of the HTTP status code.
              Successful == 200.<br>
            When HttpStatus != 200:<br>
            - HttpText: containing a string representation of the HTTP
              entity body.<br>
            When HttpText is a standard HDP WS error entity body the
            following key/values (convenience to avoid parsing
            HttpStatus):<br>
            - ErrorMessage: value from HttpText.<br>
            - ErrorDescription: value from HttpText.<br>
        """
        resource_path = ';'.join(type.value for type in resource_types)
        url = self._endpoint + 'Spase/' + resource_path

        if date_range is not None:

            if isinstance(date_range, list):
                date_interval = DateInterval(date_range[0], date_range[1])
            else:
                date_interval = date_range

            url += '/' + str(date_interval)

        response = self._session.get(url, params=query,
                                     timeout=self._timeout)

        self.logger.debug('response.url = %s', response.url)

        status = self.__get_status(response)
        if response.status_code != 200:
            return status

        result = {
            'Result': ET.fromstring(response.text)
        }

        result.update(status)
        return result


    @staticmethod
    def __get_status(
            response: requests.Response
        ) -> Dict:
        """
        Gets status information from the given response.  In particular,
        when status_code != 200, an attempt is made to extract the HDP WS
        ErrorMessage and ErrorDescription from the response.

        Parameters
        ----------
        response
            requests Response object.

        Returns
        -------
        Dict
            Dict containing the following:<br>
            - HttpStatus: the HTTP status code<br>
            additionally, when HttpStatus != 200<br>
            - ErrorText: a string representation of the entire entity
              body<br>
            - ErrorMessage: HDP WS ErrorMessage (when available)<br>
            - ErrorDescription: HDP WS ErrorDescription (when available)
        """
        http_result = {
            'HttpStatus': response.status_code
        }

        if response.status_code != 200:

            http_result['ErrorText'] = response.text
            try:
                error_element = ET.fromstring(response.text)
                http_result['ErrorMessage'] = error_element.findall(\
                    './/xhtml:p[@class="ErrorMessage"]/xhtml:b',
                    namespaces=NAMESPACES)[0].tail
                http_result['ErrorDescription'] = error_element.findall(\
                    './/xhtml:p[@class="ErrorDescription"]/xhtml:b',
                    namespaces=NAMESPACES)[0].tail
            except ParseError:
                pass  # ErrorText is the best we can do

        return http_result



#!/usr/bin/env python3

#
# NOSA HEADER START
#
# The contents of this file are subject to the terms of the NASA Open
# Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may
# not use this file except in compliance with the Agreement.
#
# You can obtain a copy of the agreement at
#   docs/NASA_Open_Source_Agreement_1.3.txt
# or
#   https://cdaweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
#
# See the Agreement for the specific language governing permissions
# and limitations under the Agreement.
#
# When distributing Covered Code, include this NOSA HEADER in each
# file and include the Agreement file at
# docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the
# following below this NOSA HEADER, with the fields enclosed by
# brackets "[]" replaced with your own identifying information:
# Portions Copyright [yyyy] [name of copyright owner]
#
# NOSA HEADER END
#
# Copyright (c) 2023 United States Government as represented by
# the National Aeronautics and Space Administration. No copyright is
# claimed in the United States under Title 17, U.S.Code. All Other
# Rights Reserved.
#

"""
Package for accessing the NASA's Heliophysics Data Portal (HDP) web
services https://heliophysicsdata.gsfc.nasa.gov/WebServices/.

Copyright &copy; 2023 United States Government as represented by the
National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S.Code. All Other Rights Reserved.

"""


__version__ = "0.4.1"


#
# Limit on the number of times an HTTP request which returns a
# 429 or 503 status with a Retry-After header will be retried.
#
#RETRY_LIMIT = 100


#
# XML schema namespace
#
HDP_NS = 'http://heliophysicsdata.gsfc.nasa.gov/hdp'
#
# SPASE schema namespace
#
SPASE_NS = 'http://www.spase-group.org/data/schema'
#
# XHTML schema namespace
#
XHTML_NS = 'http://www.w3.org/1999/xhtml'
#
# All namespaces found in responses with spase being the default.
#
NAMESPACES = {
    'hdp': HDP_NS,
    '': SPASE_NS,
    'xhtml': XHTML_NS
}



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="hdpws",
    version="0.4.1",
    description="NASA's Heliophysics Data Portal Web Service Client Library",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://heliophysicsdata.gsfc.nasa.gov/WebServices",
    author="Bernie Harris",
    author_email="NASA-SPDF-Support@nasa.onmicrosoft.com",
    license="NOSA",
    packages=["hdpws"],
#    packages=find_packages(exclude=["tests"]),
#    packages=find_packages(),
    include_package_data=True,
    install_requires=["python-dateutil>=2.8.0", "requests>=2.20"],
#    extras_require={
#        'plot': ["matplotlib>=3.3.2"],
#    },
    keywords=["heliophysics", "spase", "space physics", "spdf", "hdp"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
#        "License :: OSI Approved :: NASA Open Source Agreement",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
#    entry_points={
#        "console_scripts": [
#            hdpcws=hdpws.__main__:example",
#        ]
#    },
)



#!/usr/bin/env python3

#
# NOSA HEADER START
#
# The contents of this file are subject to the terms of the NASA Open
# Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may
# not use this file except in compliance with the Agreement.
#
# You can obtain a copy of the agreement at
#   docs/NASA_Open_Source_Agreement_1.3.txt
# or
#   https://cdaweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
#
# See the Agreement for the specific language governing permissions
# and limitations under the Agreement.
#
# When distributing Covered Code, include this NOSA HEADER in each
# file and include the Agreement file at
# docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the
# following below this NOSA HEADER, with the fields enclosed by
# brackets "[]" replaced with your own identifying information:
# Portions Copyright [yyyy] [name of copyright owner]
#
# NOSA HEADER END
#
# Copyright (c) 2023 United States Government as represented by
# the National Aeronautics and Space Administration. No copyright is
# claimed in the United States under Title 17, U.S.Code. All Other
# Rights Reserved.
#


"""
Module defining a class to represent date interval.<br>

Copyright &copy; 2023 United States Government as represented by the
National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S.Code. All Other Rights Reserved.
"""


from datetime import date
from typing import Tuple, Union



class DateInterval:
    """
    A class representing a date interval.

    Parameters
    ----------
    start
        Start date of interval.
    end
        End date of interval.

    Raises
    ------
    ValueError
        If the given start/end datedate values are invalid.
    """
    def __init__(self,
                 start: Union[date, str],
                 end: Union[date, str]):

        self._start = DateInterval.get_date(start)
        self._end = DateInterval.get_date(end)


    @property
    def start(self) -> date:
        """
        Gets the start value.

        Returns
        -------
        date
            start value.
        """
        return self._start


    @start.setter
    def start(self, value: Union[date, str]):
        """
        Sets the start value.

        Parameters
        ----------
        value
            start date value.
        """
        self._start = DateInterval.get_date(value)


    @property
    def end(self) -> date:
        """
        Gets the end value.

        Returns
        -------
        date
            end value.
        """
        return self._end


    @end.setter
    def end(self, value: Union[date, str]):
        """
        Sets the _end value.

        Parameters
        ----------
        value
            end date value.
        """
        self._end = DateInterval.get_date(value)


    def __str__(self):
        return self._start.isoformat() + ',' + self._end.isoformat()


    def __eq__(self, other):
        return self._start == other.start and self._end == other.end



    @staticmethod
    def get_date(value: Union[date, str]) -> date:
        """
        Produces a date representation of the given value. 

        Parameters
        ----------
        value
            value to convert to a date.
        Returns
        -------
        date
            date representation of the given value.
        Raises
        ------
        ValueError
            If the given value is not a valid datetime.date value.
        """
        if isinstance(value, date):
            date_value = value
        elif isinstance(value, str):
            date_value = date.fromisoformat(value)
        else:
            raise ValueError('unrecognized datetime.date value')

        return date_value


    @staticmethod
    def get_dates(
            start: Union[date, str],
            end: Union[date, str]
        ) -> Tuple[date, date]:
        """
        Produces a date representation of the given values.

        Parameters
        ----------
        start
            start value to convert to a date.
        end
            end value to convert to a date.
        Returns
        -------
        Tuple
            [0] date representation of the given start value.
            [1] date representation of the given end value.
        Raises
        ------
        ValueError
            If either of the given values is not a valid date value.
        """
        try:
            start_date = DateInterval.get_date(start)
        except ValueError as invalid_start_date:
            raise ValueError('unrecognized start date value')\
            from invalid_start_date

        try:
            end_date = DateInterval.get_date(end)
        except ValueError as invalid_end_date:
            raise ValueError('unrecognized end date value')\
            from invalid_end_date

        return start_date, end_date



#!/usr/bin/env python3

#
# NOSA HEADER START
#
# The contents of this file are subject to the terms of the NASA Open
# Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may
# not use this file except in compliance with the Agreement.
#
# You can obtain a copy of the agreement at
#   docs/NASA_Open_Source_Agreement_1.3.txt
# or
#   https://cdaweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
#
# See the Agreement for the specific language governing permissions
# and limitations under the Agreement.
#
# When distributing Covered Code, include this NOSA HEADER in each
# file and include the Agreement file at
# docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the
# following below this NOSA HEADER, with the fields enclosed by
# brackets "[]" replaced with your own identifying information:
# Portions Copyright [yyyy] [name of copyright owner]
#
# NOSA HEADER END
#
# Copyright (c) 2023 United States Government as represented by
# the National Aeronautics and Space Administration. No copyright is
# claimed in the United States under Title 17, U.S.Code. All Other
# Rights Reserved.
#

"""
Example Heliophysics Data Portal (HDP) web services client.
https://heliophysicsdata.gsfc.nasa.gov/WebServices/.

Copyright &copy; 2023 United States Government as represented by the
National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S.Code. All Other Rights Reserved.
"""

import sys
import getopt
import json
import logging
import logging.config
from typing import Dict, List
import xml.etree.ElementTree as ET  # pylint: disable=unused-import
import urllib3


from hdpws.hdpws import HdpWs
from hdpws import NAMESPACES
from hdpws.resourcetype import ResourceType as rt


logging.basicConfig()
LOGGING_CONFIG_FILE = 'logging_config.json'
try:
    with open(LOGGING_CONFIG_FILE, 'r', encoding='utf-8') as fd:
        logging.config.dictConfig(json.load(fd))
except BaseException as exc:    # pylint: disable=broad-except
    if not isinstance(exc, FileNotFoundError):
        print('Logging configuration failed')
        print('Exception: ', exc)
        print('Ignoring failure')
        print()


#ENDPOINT = "https://heliophysicsdata.gsfc.nasa.gov/WS/hdp/1/"
#ENDPOINT = "http://heliophysicsdata-dev.sci.gsfc.nasa.gov/WS/hdp/1/"
#ENDPOINT = "http://localhost:8100/exist/restxq/"
ENDPOINT = "http://localhost:8080/exist/restxq/"
#CA_CERTS = '/etc/pki/ca-trust/extracted/openssl/ca-bundle.trust.crt'


def print_usage(
        name: str
    ) -> None:
    """
    Prints program usage information to stdout.

    Parameters
    ----------
    name
        name of this program

    Returns
    -------
    None
    """
    print(f'USAGE: {name} [-e url][-d][-c cacerts][-h]')
    print('WHERE: url = HDP web service endpoint URL')
    print('       -d disables TLS server certificate validation')
    print('       cacerts = CA certificate filename')


# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def example(
        argv: List[str]
    ) -> None:
    """
    Example Heliophysics Data Portal (HDP) web service client.
    Includes example calls to most of the web services.

    Parameters
    ----------
    argv
        Command-line arguments.<br>
        -e url or --endpoint=url where url is the cdas web service endpoint
            URL to use.<br>
        -c url or --cacerts=filename where filename is the name of the file
            containing the CA certificates to use.<br>
        -d or --disable-cert-check to disable verification of the server's
            certificate
        -h or --help prints help information.
    """

    try:
        opts = getopt.getopt(argv[1:], 'he:c:d',
                             ['help', 'endpoint=', 'cacerts=',
                              'disable-cert-check'])[0]
    except getopt.GetoptError:
        print('ERROR: invalid option')
        print_usage(argv[0])
        sys.exit(2)

    endpoint = ENDPOINT
    ca_certs = None
    disable_ssl_certificate_validation = False

    for opt, arg in opts:
        if opt in ('-e', '--endpoint'):
            endpoint = arg
        elif opt in ('-c', '--cacerts'):
            ca_certs = arg
        elif opt in ('-d', '--disable-cert-check'):
            disable_ssl_certificate_validation = True
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        elif opt in ('-h', '--help'):
            print_usage(argv[0])
            sys.exit()

    hdp = HdpWs(endpoint=endpoint, ca_certs=ca_certs,
                disable_ssl_certificate_validation=
                disable_ssl_certificate_validation, user_agent='Example')


    result = hdp.get_instrument_ids()
    if result['HttpStatus'] == 200:
        print('HDP InstrumentIDs:')
        for value in result['InstrumentID'][0:5]:
            print(f'    {value}')
        print('    ...')
    else:
        print_error('hdp.get_instrument_ids', result)

    result = hdp.get_repository_ids()
    if result['HttpStatus'] == 200:
        print('HDP RepositoryIDs:')
        for value in result['RepositoryID'][0:5]:
            print(f'    {value}')
        print('    ...')
    else:
        print_error('hdp.get_repository_ids', result)

    result = hdp.get_measurement_types()
    if result['HttpStatus'] == 200:
        print('HDP MeasurementTypes:')
        for value in result['MeasurementType']:
            print(f'    {value}')
    else:
        print_error('hdp.get_measurement_types', result)

    result = hdp.get_spectral_ranges()
    if result['HttpStatus'] == 200:
        print('HDP SpectralRanges:')
        for value in result['SpectralRange']:
            print(f'    {value}')
    else:
        print_error('hdp.get_spectral_ranges', result)

    result = hdp.get_phenomenon_types()
    if result['HttpStatus'] == 200:
        print('HDP PhenomenonType:')
        for value in result['PhenomenonType']:
            print(f'    {value}')
    else:
        print_error('hdp.get_phenomenon_types', result)

    result = hdp.get_observatory_group_ids()
    if result['HttpStatus'] == 200:
        print('HDP ObservatoryGroupIDs:')
        for value in result['ObservatoryGroupID'][0:5]:
            print(f'    {value}')
        print('    ...')
    else:
        print_error('hdp.get_observatory_group_ids', result)

    result = hdp.get_observatory_ids()
    if result['HttpStatus'] == 200:
        print('HDP ObservatoryIDs:')
        for value in result['ObservatoryID'][0:5]:
            print(f'    {value}')
        print('    ...')
    else:
        print_error('hdp.get_observatory_group_ids', result)

    result = hdp.get_observed_regions()
    if result['HttpStatus'] == 200:
        print('HDP Observed Regions:')
        for value in result['ObservedRegion'][0:5]:
            print(f'    {value}')
        print('    ...')
    else:
        print_error('hdp.get_observed_regions', result)


    resource_ids = [
        'spase://NASA/Collection/IRIS_AIA',
        'spase://SMWG/Service/CCMC/Models'
    ]
    result = hdp.get_spase(resource_ids)
    if result['HttpStatus'] == 200:
        #print('HDP Spase:')
        #print(ET.tostring(result['Result']))
        print('get_spase Result ResourceIDs:')
        for spase in result['Result'].findall('.//Spase', namespaces=NAMESPACES):
            print(spase.findall('.//ResourceID', namespaces=NAMESPACES)[0].text)
            print('    ', spase.findall('.//ResourceName', namespaces=NAMESPACES)[0].text)
    else:
        print_error('hdp.get_spase_data', result)


    query = {
        'ResourceID': ['spase://NASA/NumericalData/ACE/CRIS/L2/P1D',
            'spase://NASA/NumericalData/ACE/CRIS/L2/PT1H'
        ],
#        'InstrumentID': 'spase://SMWG/Instrument/ACE/CRIS',
#        'ObservatoryID': 'spase://SMWG/Observatory/ACE',
#        'Cadence': '=PT1H',
#        'ObservedRegion': 'Heliosphere.NearEarth',
#        'MeasurementType': 'EnergeticParticles',
#        'AccessRights': 'Open',
#        'Format': 'CDF'
    }
    types = [rt.NUMERICAL_DATA, rt.DISPLAY_DATA]
    time_range = ['2022-01-01', '2022-01-02']

    result = hdp.get_spase_data(types, query, time_range)
    if result['HttpStatus'] == 200:
        #print('HDP Spase:')
        #print(ET.tostring(result['Result']))
        print('get_spase_data Result ResourceIDs:')
        for r_id in result['Result'].findall('.//ResourceID', namespaces=NAMESPACES):
            print(r_id.text)
    else:
        print_error('hdp.get_spase_data', result)


    query = {
        'ResourceID': ['spase://NASA/NumericalData/ACE/MAG/L2/PT16S', 'bad']
    }

    result = hdp.get_spase_data(types, query)
    if result['HttpStatus'] == 200:
        print('get_spase_data Result:')
        print(result['Result'])
    else:
        print_error('hdp.get_spase_data', result)


    query = {
        'InstrumentID': 'spase://SMWG/Instrument/ACE/MAG',
        'PhenomenonType': 'MagneticCloud',
        'Description': 'ICME'
    }
    time_range = ['1999-01-01', '1999-01-02']

    result = hdp.get_spase_catalog(query, time_range)
    if result['HttpStatus'] == 200:
        #print('HDP Spase:')
        #print(ET.tostring(result['Result']))
        print('Result Catalogs:')
        for catalog in result['Result'].findall('.//Catalog', namespaces=NAMESPACES):
            print(catalog.findall('.//ResourceID', namespaces=NAMESPACES)[0].text)
            print('    ', catalog.findall('.//ResourceName', namespaces=NAMESPACES)[0].text)
            #ET.indent(catalog)
            #print(ET.tostring(catalog, encoding='unicode',
            #                  default_namespace='http://www.spase-group.org/data/schema'))
    else:
        print_error('hdp.get_spase_catalog', result)


def print_error(
        label: str,
        result: Dict
    ) -> None:
    """
    Prints an error result.

    Parameters
    ----------
    result
        Dictionary result returned by the hdpws.
    """
    print(f'{label} failed with status = {result["HttpStatus"]}')
    if 'ErrorMessage' in result:
        print(f'ErrorMessage = {result["ErrorMessage"]}')
        print(f'ErrorDescription = {result["ErrorDescription"]}')
    else:
        print(f'HttpText = {result["ErrorText"]}')


if __name__ == '__main__':
    example(sys.argv)
